#CALCOFI_CONVERSION.PL
#
#This script converts geographic coordinates (latitude/longitude) to CalCOFI station grid coordinates (line and station numbers)
#or vice versa.
#
#A conversion flag identifies the direction of the conversion.
#

#! /usr/bin/perl -w
#
use strict;

######################## USER-MODIFIED SETTINGS
our $input_fname = "coords.in";
our $output_fname = "coords.out";

#this flag identifies the direction of the conversion
#a value of "geo2cc" indicates from geographic coordinates (latitude/longitude) to CalCOFI station grid coordinates
#a value of "cc2geo" indicates from CalCOFI station grid coordinates to geographic coordinates (latitude/longitude)
our $conversion = "geo2cc";
#our $conversion = "cc2geo";
########################

open(INFILE, $input_fname) || die "Cannot open INPUT_FILE filehandle: $!\n";
open(OUTFILE, ">" . $output_fname) || die "Cannot open OUTPUT_FILE filehandle: $!\n";

if ($conversion eq "geo2cc") {
	print "Converting data in $input_fname from geographic coordinates to CalCOFI station grid coordinates...\n";
}
if ($conversion eq "cc2geo") {
	print "Converting data in $input_fname from CalCOFI station grid coordinates to geographic coordinates...\n";
}

print "\nReading data...\n";
our @inputrecs=<INFILE>;
chomp(@inputrecs);
close(INFILE);

#process lat/long data and calculate CalCOFI station-line grid coordinates; format other data and print all to file
our $m;
our $line;
our $station;
our $latd;
our $latm;
our $lats;
our $lath;
our $lond;
our $lonm;
our $lons;
our $lonh;
our $latdecdeg;
our $londecdeg;

#print header records in output file
if ($conversion eq "geo2cc") {
	print OUTFILE "LATD,LATM,LATS,LATH,LOND,LONM,LONS,LONH,LAT_DD,LON_DD,LINE,STATION\n";
}
elsif ($conversion eq "cc2geo") {
        print OUTFILE "LINE,STATION,LAT_DD,LON_DD\n";
}
else {
	print "Do not recognize the selected conversion!  Should be either geo2cc or cc2geo.  Please correct.\n";
	print "Press any key and program will exit...\n";
	our $abort = <STDIN>;
	exit;
}

print "\nProcessing data and writing to output file...\n";

#skip header record in the input file
for ($m=1; $m<=$#inputrecs; $m++) {
	if ($conversion eq "geo2cc") {

		($latd,$latm,$lats,$lath,$lond,$lonm,$lons,$lonh) = split(/,/, $inputrecs[$m]);

		#get CC line and station numbers
		($line, $station, $latdecdeg, $londecdeg) = &ll2cc($latd,$latm,$lats,$lath,$lond,$lonm,$lons,$lonh);

		#round CC line and station numbers to 4 decimal places
		$line = sprintf( "%.4f", $line);
		$station = sprintf( "%.4f", $station);

		print OUTFILE "$latd,$latm,$lats,$lath,$lond,$lonm,$lons,$lonh,$latdecdeg,$londecdeg,$line,$station\n";

	}
	if ($conversion eq "cc2geo") {

		($line,$station) = split(/,/, $inputrecs[$m]);

		#get latitude and longitude values
		($line, $station, $latdecdeg, $londecdeg) = &cc2ll($line,$station);

		#round latitude and longitude to 4 decimal places
		$latdecdeg = sprintf( "%.4f", $latdecdeg);
		$londecdeg = sprintf( "%.4f", $londecdeg);

		print OUTFILE "$line,$station,$latdecdeg,$londecdeg\n";

	}
}

close(OUTFILE);

print "\nConverted data written to $output_fname ... program finished!\n";

sub ll2cc {
	#Subroutine to implement algorithm that converts lat/long coordinates to CalCOFI grid coordinates (line and station).
	#
	#INPUT: Latitude and longitude values in DMS to be converted to CC grid
	#OUTPUT: A line and station value for CC grid (and signed decimal degrees of latitude and longitude)
	#
	#
	use Math::Trig;

	my $latitude_deg = shift;
	my $latitude_min = shift;
	my $latitude_sec = shift;
	my $latitude_hem = shift;
	my $longitude_deg = shift;
	my $longitude_min = shift;
	my $longitude_sec = shift;
	my $longitude_hem = shift;

	my $latdd = $latitude_deg + ($latitude_min/60.0) +  ($latitude_sec/3600.0);
	my $londd = $longitude_deg + ($longitude_min/60.0) +  ($longitude_sec/3600.0);

	if ($latitude_hem eq "S" || $latitude_hem eq "s") {
		$latdd = $latdd*(-1.0);
	}

	if ($longitude_hem eq "W" || $longitude_hem eq "w") {
		$londd = $londd*(-1.0);
	}

	#Note:  The following is the reverse of typical conventions for E/W longitude and sign.
	#The CalCOFI grid conversion equations assume longitude is positive value for western hemisphere.
	#
	my $pla = $latdd;
	my $plo = $londd;
	$plo = $plo*(-1.0);
	
	my $L1 = ( &mctr($pla) -  &mctr(34.15) ) * tan(deg2rad(30));
	my $L2 = $plo - $L1 - 121.15;
	my $mctr_rla_val = ( $L2 * cos(deg2rad(30)) * sin(deg2rad(30)) ) + &mctr($pla);
	my $rla = &invmctr($mctr_rla_val);
	my $pln = 80.0 - ((($rla - 34.15) * 5.0)/cos(deg2rad(30)));
	my $psn = 60.0 + ((($rla - $pla) * 15.0)/sin(deg2rad(30)));

	return ($pln, $psn, $latdd, $londd);
}

sub cc2ll {
	#Subroutine to implement algorithm that converts CalCOFI grid coordinates (line and station) to lat/long coordinates.
	#
	#INPUT: Line and station value(s) for CC grid to be converted to geographic coordinates
	#OUTPUT: Latitude and longitude values in decimal degrees (signed).
	#
	#
	use Math::Trig;

	my $pln = shift;
	my $psn = shift;
	
	my $rla = 34.15 - (0.2 * ($pln - 80.0) * cos(deg2rad(30)));
	my $pla = $rla - ((1.0/15.0) * ($psn - 60.0) * sin(deg2rad(30)));
	my $L1 = ( &mctr($pla) -  &mctr(34.15) ) * tan(deg2rad(30));
	my $L2 = ( &mctr($rla) -  &mctr($pla) ) / (cos(deg2rad(30)) * sin(deg2rad(30)));
	my $plo = $L1 + $L2 + 121.15;

	#Note:  The CalCOFI grid conversion equations assume longitude is positively signed in the western hemisphere.
	#The following changes the sign on the longitude value to be in the convention of negatively signed in the western hemisphere.
	#
	$plo = $plo * (-1.0);

	return ($pln, $psn, $pla, $plo);
}

sub mctr {
	#Subroutine to implement mercator transform function.
	#
	#INPUT: A latitude value in decimal degrees.
	#OUPUT: A value in "mercator meridional units"
	#
	#
	use Math::Trig;
	my $deg = shift;
	my $mctr_val = (180.0/pi) * ( log(tan(deg2rad(45.0 + ($deg/2.0)))) - (0.00676866 * sin(deg2rad($deg))) );

	return $mctr_val;
}

sub invmctr {
	#Subroutine to implement inverse mercator transform function.
	#Because this function does not have a precise algebraic form, an
	#iterative process is employed.
	#
	#INPUT: A value in "mercator meridional units"
	#OUTPUT:  A value in decimal degrees
	#
	#
	my $i;
	my $mctr_rla_val = shift;
	my $rla = $mctr_rla_val;
	my $rla_init = $rla;
	use Math::Trig;
	for ($i=0; $i < 5; $i++) {
		$rla = 2 * ( (180/pi) * atan( exp(deg2rad($rla_init) + (0.00676866*sin(deg2rad($rla))))   ) - 45);
	}

	return $rla;
}
